from __future__ import annotations

import json
import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from string import ascii_lowercase
from typing import Dict, List, Tuple, Type, Union

import pcbnew

from .board_modifier import KICAD_VERSION
from .kle_serial import Keyboard, MatrixAnnotatedKeyboard, get_keyboard

logger = logging.getLogger(__name__)


@dataclass
class Footprint:
    libname: str
    name: str

    @classmethod
    def from_str(cls: Type[Footprint], arg: str) -> Footprint:
        parts = arg.rsplit(":", 1)
        if len(parts) != 2:
            msg = f"Unexpected footprint value: `{arg}`"
            raise RuntimeError(msg)
        return Footprint(libname="".join(parts[0]), name=parts[1])

    def load(self) -> pcbnew.FOOTPRINT:
        return pcbnew.FootprintLoad(self.libname, self.name)


class BoardBuilder:
    def __init__(
        self,
        *,
        switch_footprint: str,
        diode_footprint: str,
    ) -> None:
        self.switch_footprint = Footprint.from_str(switch_footprint)
        self.diode_footprint = Footprint.from_str(diode_footprint)

        self.board = pcbnew.CreateEmptyBoard()
        self.nets: dict[str, pcbnew.NETINFO_ITEM] = {}
        self.net_info = self.board.GetNetInfo()
        self.net_count = self.board.GetNetCount()

    def add_footprint(self, footprint: pcbnew.FOOTPRINT) -> pcbnew.FOOTPRINT:
        logger.info(f"Add {footprint.GetReference()} footprint")
        self.board.Add(footprint)
        return footprint

    def add_switch_footprint(self, ref: str) -> pcbnew.FOOTPRINT:
        f = self.switch_footprint.load()
        f.SetReference(ref)
        return self.add_footprint(f)

    def add_diode_footprint(self, ref: str) -> pcbnew.FOOTPRINT:
        f = self.diode_footprint.load()
        f.SetReference(ref)
        return self.add_footprint(f)

    def add_net(self, netname: str) -> pcbnew.NETINFO_ITEM:
        """Add new net with netname if it does not exist already"""
        if netname in self.nets:
            net = self.nets[netname]
        else:
            net = pcbnew.NETINFO_ITEM(self.board, netname, self.net_count)
            if KICAD_VERSION < (8, 0, 0):
                self.net_info.AppendNet(net)
            logger.info(f"Add {netname} net")
            self.board.Add(net)
            self.net_count += 1
            self.nets[netname] = net
        return net

    def create_board(
        self,
        keyboard: Union[str, MatrixAnnotatedKeyboard],
    ) -> pcbnew.BOARD:
        if isinstance(keyboard, str):
            with open(keyboard, "r", encoding="utf-8") as f:
                layout = json.load(f)
                tmp: Keyboard = get_keyboard(layout)
                if not isinstance(tmp, MatrixAnnotatedKeyboard):
                    try:
                        _keyboard = MatrixAnnotatedKeyboard(tmp.meta, tmp.keys)
                    except Exception as e:
                        msg = (
                            f"Layout from {keyboard} is not convertible to "
                            "matrix annotated keyboard which is required "
                            "for board create"
                        )
                        raise RuntimeError(msg) from e
                else:
                    _keyboard = tmp
        else:
            _keyboard: MatrixAnnotatedKeyboard = keyboard

        _keyboard.collapse()

        current_ref = 1
        items: List[Tuple[str, str]] = []
        key_iterator = _keyboard.key_iterator(ignore_alternative=False)

        for key in key_iterator:
            if key.decal:
                continue
            items.append(MatrixAnnotatedKeyboard.get_matrix_position(key))

        def _sort_matrix(item: Tuple[str, str]) -> Tuple[int, int]:
            row_match = re.search(r"\d+", item[0])
            column_match = re.search(r"\d+", item[1])

            if row_match is None or column_match is None:
                msg = f"No numeric part for row or column found in '{item}'"
                raise ValueError(msg)

            return int(row_match.group()), int(column_match.group())

        progress: Dict[Tuple[str, str], List[pcbnew.FOOTPRINT]] = defaultdict(list)
        for row, column in sorted(items, key=_sort_matrix):
            position = (row, column)
            if position not in progress:
                switch = self.add_switch_footprint(f"SW{current_ref}")
                diode = self.add_diode_footprint(f"D{current_ref}")

                column_name = f"COL{column}" if column.isdigit() else column
                net = self.add_net(column_name)
                switch.FindPadByNumber("1").SetNet(net)

                row_name = f"ROW{row}" if row.isdigit() else row
                net = self.add_net(row_name)
                diode.FindPadByNumber("1").SetNet(net)

                net = self.add_net(f"Net-(D{current_ref})-Pad2")
                switch.FindPadByNumber("2").SetNet(net)
                diode.FindPadByNumber("2").SetNet(net)

                current_ref += 1
                progress[position].append(switch)
            else:
                # this must be alternative layout key, do not need to create nets
                # or add diode for it. Just add duplicate and add suffix to reference
                switches = progress[position]
                default_switch = switches[0]

                suffix = ascii_lowercase[len(switches) - 1]
                switch = pcbnew.Cast_to_FOOTPRINT(default_switch.Duplicate())
                switch.SetReference(default_switch.GetReference() + suffix)
                self.add_footprint(switch)

                progress[position].append(switch)

        return self.board



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class Side(str, Enum):
    FRONT = "Front"
    BACK = "Back"

    @classmethod
    def get(cls, name: str) -> Side:
        if isinstance(name, str):
            try:
                return Side(name.title())
            except ValueError:
                # fallback to error below to use 'name' before converting to titlecase
                pass
        msg = f"'{name}' is not a valid Side"
        raise ValueError(msg)


@dataclass
class ElementPosition:
    x: float
    y: float
    orientation: float
    side: Side


class PositionOption(str, Enum):
    DEFAULT = "Default"
    RELATIVE = "Relative"
    PRESET = "Preset"
    CUSTOM = "Custom"
    UNCHANGED = "Unchanged"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def get(cls, name: str) -> PositionOption:
        if isinstance(name, str):
            try:
                return PositionOption(name.title())
            except ValueError:
                # fallback to error below to use 'name' before converting to titlecase
                pass
        msg = f"'{name}' is not a valid PositionOption"
        raise ValueError(msg)


@dataclass
class ElementInfo:
    annotation_format: str
    position_option: PositionOption
    position: Optional[ElementPosition]
    template_path: str

    @classmethod
    def from_dict(cls, data: dict) -> ElementInfo:
        position_data = data.pop("position", None)
        position = ElementPosition(**position_data) if position_data else None
        return cls(position=position, **data)



import argparse
import logging
import os
import sys
from typing import List

import pcbnew

from . import __version__
from .defaults import DEFAULT_DIODE_POSITION, ZERO_POSITION
from .element_position import ElementInfo, ElementPosition, PositionOption, Side
from .kbplacer_plugin import PluginSettings, run

logger = logging.getLogger(__name__)


def check_annotation(value: str) -> None:
    if value.count("{}") != 1:
        err = (
            f"'{value}' invalid annotation specifier, "
            "it must contain exactly one '{}' placeholder."
        )
        raise ValueError(err)


class SwitchElementInfoAction(argparse.Action):
    """Simplified action producing ElementInfo which always use
    `PositionOption.DEFAULT` and `ElementPosition` with x and y 0, i.e. the only
    variable (configurable) parameters are annotation, orientation and side
    """

    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value: ElementInfo = self.parse(values, option_string)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)

    def parse(self, values: str, option_string) -> ElementInfo:
        tokens: list[str] = values.split()
        err = ""
        if len(tokens) not in [1, 3]:
            err = f"{option_string} invalid format."
            raise ValueError(err)

        annotation = tokens[0]
        check_annotation(annotation)

        if len(tokens) > 1:
            orientation = float(tokens[1])
            side = Side.get(tokens[2])
        else:
            orientation = 0
            side = Side.FRONT

        return ElementInfo(
            annotation,
            PositionOption.DEFAULT,
            ElementPosition(0, 0, orientation, side),
            "",
        )


class ElementInfoAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value: ElementInfo = self.parse(values, option_string)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)

    def parse(self, values: str, option_string) -> ElementInfo:
        tokens: list[str] = values.split()
        err = ""
        if len(tokens) not in [2, 3, 6]:
            err = f"{option_string} invalid format."
            raise ValueError(err)

        annotation = tokens[0]
        check_annotation(annotation)

        option = PositionOption.get(tokens[1])
        position = None
        template_path = ""

        if len(tokens) == 2:
            if option not in [
                PositionOption.RELATIVE,
                PositionOption.DEFAULT,
                PositionOption.UNCHANGED,
            ]:
                err = (
                    f"{option_string} position option needs to be equal "
                    "RELATIVE or DEFAULT if position details not provided"
                )
                raise ValueError(err)
        elif len(tokens) == 3:
            if option not in [PositionOption.PRESET, PositionOption.RELATIVE]:
                err = (
                    f"{option_string} position option needs to be equal "
                    "RELATIVE or PRESET when providing template path"
                )
                raise ValueError(err)
        elif option != PositionOption.CUSTOM:
            err = (
                f"{option_string} position option needs to be equal CUSTOM "
                "when providing position details"
            )
            raise ValueError(err)

        if option == PositionOption.CUSTOM:
            floats = tuple(map(float, tokens[2:5]))
            side = Side.get(tokens[5])
            position = ElementPosition(floats[0], floats[1], floats[2], side)
        elif option == PositionOption.RELATIVE:
            # template path is optional for RELATIVE option:
            if len(tokens) == 3:
                template_path = tokens[2]
        elif option == PositionOption.PRESET:
            template_path = tokens[2]

        value: ElementInfo = ElementInfo(
            annotation,
            option,
            position,
            template_path,
        )
        return value


class ElementInfoListAction(ElementInfoAction):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            if "DEFAULT" in values:
                msg = f"{option_string} does not support DEFAULT position"
                raise ValueError(msg)
            value: List[ElementInfo] = []
            tokens: list[str] = values.split(";")
            for token in tokens:
                element_info: ElementInfo = super().parse(token.strip(), option_string)
                value.append(element_info)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


class XYAction(argparse.Action):
    def __call__(self, parser, namespace, values: str, option_string=None) -> None:
        try:
            value = tuple(map(float, values.split()))
            if len(value) != 2:
                msg = (
                    f"{option_string} must be exactly two numeric values "
                    "separated by a space."
                )
                raise ValueError(msg)
        except ValueError as e:
            raise argparse.ArgumentTypeError(str(e))
        setattr(namespace, self.dest, value)


def app() -> None:
    parser = argparse.ArgumentParser(
        description="Keyboard's key autoplacer",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "-b",
        "--board",
        required=True,
        help=".kicad_pcb file to be processed or created",
    )
    parser.add_argument(
        "-l", "--layout", required=False, default="", help="json layout definition file"
    )
    parser.add_argument(
        "--route-switches-with-diodes",
        action="store_true",
        help="Enable switch-diode routing",
    )
    parser.add_argument(
        "--route-rows-and-columns",
        action="store_true",
        help="Enable rows/columns routing",
    )
    parser.add_argument(
        "-s",
        "--switch",
        default=ElementInfo("SW{}", PositionOption.DEFAULT, ZERO_POSITION, ""),
        action=SwitchElementInfoAction,
        help=(
            "Switch information, space separated value of ANNOTATION [ORIENTATION SIDE]\n"
            "where ANNOTATION is footprint annotation string with '{}' placeholder,\n"
            "ORIENTATION is numeric value of reference orientation for each footprint\n"
            "and SIDE is either FRONT or BACK\n"
            "When only ANNOTATION defined, then orientation is 0 and side is FRONT\n"
            "for example:\n"
            "\tSW{}\n"
            "\tSW{} 0 FRONT\n"
            "\tSW{} 90 BACK\n"
            "equal 'SW{} 0 FRONT' by default"
        ),
    )
    parser.add_argument(
        "-d",
        "--diode",
        default=ElementInfo("D{}", PositionOption.DEFAULT, DEFAULT_DIODE_POSITION, ""),
        action=ElementInfoAction,
        help=(
            "Diode information, space separated value of ANNOTATION OPTION [POSITION]\n"
            "Available OPTION choices: DEFAULT, UNCHANGED, RELATIVE, PRESET and CUSTOM\n"
            "When DEFAULT or UNCHAGED, then POSITION needs to be omitted,\n"
            "when RELATIVE, then POSITION is optional path for saving kicad_pcb template file\n"
            "when PRESET, then POSITION is mandatory path to kicad_pcb template file\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tD{} RELATIVE\n"
            "\tD{} PRESET /home/user/project/diode_preset.kicad_pcb\n"
            "\tD{} CUSTOM 5 -4.5 90 BACK\n"
            "equal 'D{} DEFAULT' by default"
        ),
    )
    parser.add_argument(
        "--additional-elements",
        default=[
            ElementInfo(
                "ST{}",
                PositionOption.CUSTOM,
                ZERO_POSITION,
                "",
            )
        ],
        action=ElementInfoListAction,
        help=(
            "List of ';' separated additional elements ELEMENT_INFO values\n"
            "ELEMENT_INFO is space separated value of ANNOTATION OPTION POSITION\n"
            "Available OPTION choices: RELATIVE, PRESET and CUSTOM\n"
            "when RELATIVE, then POSITION is optional path for saving kicad_pcb template file\n"
            "when PRESET, then POSITION is mandatory path to kicad_pcb template file\n"
            "when CUSTOM, then POSITION is space separated value of X Y ORIENTATION FRONT|BACK\n"
            "for example:\n"
            "\tST{} CUSTOM 0 0 180 BACK;LED{} RELATIVE\n"
            "equal 'ST{} CUSTOM 0 0 0 FRONT' by default"
        ),
    )
    parser.add_argument(
        "--key-distance",
        default=(19.05, 19.05),
        action=XYAction,
        help=(
            "X and Y key 1U distance in mm, as two space separated numeric values, "
            "19.05 19.05 by default"
        ),
    )
    parser.add_argument(
        "-t",
        "--template",
        required=False,
        default="",
        help="Controller circuit template",
    )
    parser.add_argument(
        "--build-board-outline",
        required=False,
        action="store_true",
        help="Enables board outline generation around switch footprints.",
    )
    parser.add_argument(
        "--outline-delta",
        required=False,
        default=0.0,
        type=float,
        help="The amount (in millimetres) to inflate/deflate board outline.",
    )
    parser.add_argument(
        "--create-from-annotated-layout",
        required=False,
        action="store_true",
        help=(
            "Enables experimental creation mode, where via-annotated kle layout is used\n"
            "for adding footprints and netlists to the newly created board file.\n"
        ),
    )
    parser.add_argument(
        "--switch-footprint",
        required=False,
        default="",
        type=str,
        help=(
            "Full path to switch footprint, required when "
            "`--create-from-annotated-layout` option used."
        ),
    )
    parser.add_argument(
        "--diode-footprint",
        required=False,
        default="",
        type=str,
        help=(
            "Full path to diode footprint, required when "
            "`--create-from-annotated-layout` option used."
        ),
    )
    parser.add_argument(
        "--log-level",
        required=False,
        default="WARNING",
        choices=logging._nameToLevel.keys(),
        type=str,
        help="Provide logging level, default=%(default)s",
    )
    parser.add_argument(
        "--optimize-diodes-orientation",
        action="store_true",
        help=(
            "Enables automatic diode orientation adjustment based on distance\n"
            "between diode and corresponding switch."
        ),
    )

    args = parser.parse_args()

    layout_path = args.layout
    board_path = args.board

    # set up logger
    logging.basicConfig(
        level=args.log_level, format="%(asctime)s: %(message)s", datefmt="%H:%M:%S"
    )

    if args.create_from_annotated_layout and os.path.isfile(board_path):
        logger.error(f"File {board_path} already exist, aborting")
        sys.exit(1)

    settings = PluginSettings(
        board_path=board_path,
        layout_path=layout_path,
        key_info=args.switch,
        key_distance=args.key_distance,
        diode_info=args.diode,
        route_switches_with_diodes=args.route_switches_with_diodes,
        optimize_diodes_orientation=args.optimize_diodes_orientation,
        route_rows_and_columns=args.route_rows_and_columns,
        additional_elements=args.additional_elements,
        generate_outline=args.build_board_outline,
        outline_delta=args.outline_delta,
        template_path=args.template,
        create_from_annotated_layout=args.create_from_annotated_layout,
        switch_footprint=args.switch_footprint,
        diode_footprint=args.diode_footprint,
    )
    board = run(settings)

    pcbnew.Refresh()
    pcbnew.SaveBoard(board_path, board)

    logging.shutdown()


if __name__ == "__main__":
    app()



import logging

import pcbnew

from .board_modifier import (
    get_footprint,
    get_orientation,
    get_position,
    get_side,
    set_position,
    set_rotation,
    set_side,
)

logger = logging.getLogger(__name__)


def copy_from_template_to_board(
    board: pcbnew.BOARD,
    template_path: str,
    route_tracks: bool,
) -> None:
    """Copy positions of elements and tracks from template to board.
    This method does not copy parts itself - parts to be positioned
    need to be present in board prior to calling this.
    """
    template = pcbnew.LoadBoard(template_path)
    footprints = template.GetFootprints()
    board_nets_by_name = board.GetNetsByName()

    for footprint in footprints:
        reference = footprint.GetReference()
        destination_footprint = get_footprint(board, reference)

        side = get_side(footprint)
        position = get_position(footprint)
        orientation = get_orientation(footprint)

        set_side(destination_footprint, side)
        set_position(destination_footprint, position)
        set_rotation(destination_footprint, orientation)

    if route_tracks:
        tracks = template.GetTracks()
        for track in tracks:
            # Clone track but remap netinfo because net codes in template
            # might be different. Use net names for remapping
            # (names in template and board under modification must match)
            clone = track.Duplicate()
            net_name = clone.GetNetname()
            net_code = clone.GetNetCode()
            net_info_in_board = board_nets_by_name[net_name]
            logger.info(
                f"Cloning track from template: {net_name}:{net_code}"
                f"-> {net_info_in_board.GetNetname()}:"
                f"{net_info_in_board.GetNetCode()}",
            )
            clone.SetNet(net_info_in_board)
            board.Add(clone)
